/*
  This code was written by Raymond Chi-Wing Wong.
  Please read the following.

    Raymond Chi-Wing Wong, Yufei Tao, Ada Wai-Chee Fu and Xiaokui Xiao. 
    "On Efficient Spatial Matching", 
    the 33rd International Conference on Very Large Data Bases (VLDB'07), 
    Vienna, Austria on 23-28 Sept, 2007 

*/

#ifndef ASSIGN_H
#define ASSIGN_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

struct triple_s
{
	int oid;
	int pid;
	int weight;
};

typedef struct triple_s triple;

struct assignment_s
{
	int noOfMatch;

	triple **list;
};

typedef struct assignment_s assignment;

struct assignmentSet_s
{
	int noOfAssign;

	assignment **list;
};

typedef struct assignmentSet_s assignmentSet;

triple *triple_newValue(int oid, int pid, int weight);
void assignment_insert(assignment *aAssignment, triple **aTriple);

assignment *assignment_new();


void assignment_print(char *filename, assignment *aAssignment);
void assignment_removeIndex(assignment *aAssignment, int index);

assignmentSet *assignmentSet_new();
void assignmentSet_free(assignmentSet **aAssignmentSet);

assignment *assignment_union(assignment *aAssignment1, assignment *aAssignment2);

void assignment_free(assignment **aAssignment);

void assignmentSet_transformFromFinalAssignment(assignment *aAssignment, int noOfPointO, int noOfPointP, assignmentSet **assignmentSetForO, assignmentSet **assignmentSetForP);

#endif
