/*
  This code was written by Raymond Chi-Wing Wong.
  Please read the following.

    Raymond Chi-Wing Wong, Yufei Tao, Ada Wai-Chee Fu and Xiaokui Xiao. 
    "On Efficient Spatial Matching", 
    the 33rd International Conference on Very Large Data Bases (VLDB'07), 
    Vienna, Austria on 23-28 Sept, 2007 

*/

#include <stdio.h>
#include <stdlib.h>

#ifndef WIN32
#include <sys/resource.h>
#include <sys/times.h>
#include <unistd.h>
#endif

#include "rsa.h"
#include "config.h"
#include "stable.h"
#include "closestPairMatrix.h"
#include "closestPairTree.h"

statistics *globalStat;


#ifndef WIN32
void calculateExecutionTime(struct rusage *myTimeStart, struct rusage *myTimeEnd, float *userTime, float *sysTime)
{
	(*userTime) =
		((float) (myTimeEnd->ru_utime.tv_sec  - myTimeStart->ru_utime.tv_sec)) +
		((float) (myTimeEnd->ru_utime.tv_usec - myTimeStart->ru_utime.tv_usec)) * 1e-6;
	(*sysTime) =
		((float) (myTimeEnd->ru_stime.tv_sec  - myTimeStart->ru_stime.tv_sec)) +
		((float) (myTimeEnd->ru_stime.tv_usec - myTimeStart->ru_stime.tv_usec)) * 1e-6;
	
}
#endif

node_type *quota_buildRtree(pointList *aPointList, rtree_info *aInfo)
{ 
	node_type *returnValue;
	int i, j;

	int noOfPoint;
	point **pointList;
	point *aPoint;

	int pointID;

	R_TYPE *content;
	
	/* make tree root node first */
	
	tree_node_allocate(&returnValue, aInfo);
	
	for(i=0; i < aInfo->dim; i++) { 
		returnValue->b[i] = (R_TYPE)(-1 * INT_MAX);
       	returnValue->a[i] = (R_TYPE)(INT_MAX);
	}
	returnValue->id = -3;		//NO_ID
	returnValue->attribute=ROOT;
	returnValue->vacancy=aInfo->M;   
	returnValue->parent=NULL; 
	for(j=0; j < aInfo->M; j++)
		returnValue->ptr[j]=NULL;
	

	noOfPoint = aPointList->noOfPoint;
	pointList = aPointList->list;
	
	/* add data to the tree */
	for(i=0; i< noOfPoint; i++) 
	{ 		
		aPoint = pointList[i];

		pointID = aPoint->id;
		content = aPoint->content;

#ifdef DEBUG
		if (i+1%1000==0)
			printf("    insert data %d\n", pointID);
#endif

		insert_node(returnValue, content, pointID, aInfo);  
	}

	return returnValue;
} 

/*
int main()
{
	configure *aConfigure;
	char *confFilename = "conf.txt";
	char *assignFilename = "assign.txt";
	char *debugFilename = "debug.txt";

	FILE *debugFilePtr;

#ifndef WIN32
	char *timeFilename = "time.txt";
	FILE *timeFilePtr;

	float userTime_buildRtreeO, sysTime_buildRtreeO;
	float userTime_buildRtreeP, sysTime_buildRtreeP;
	float userTime_totalBuildRTree, sysTime_totalBuildRTree;
	float userTime_constructPrefList, sysTime_constructPrefList;
	float userTime_constructRankList, sysTime_constructRankList;
	float userTime_totalConstructPrefRankList, sysTime_totalConstructPrefRankList;
	float userTime_totalConstructMatrix, sysTime_totalConstructMatrix;
	float userTime_mainAlg, sysTime_mainAlg;
	float userTime_allTime, sysTime_allTime;

	struct rusage myTime_start;
	struct rusage myTime_buildRtreeO_end;
	struct rusage myTime_buildRtreeP_end;
	struct rusage myTime_constructPrefList_end;
	struct rusage myTime_constructRankList_end;
	struct rusage myTime_constructMatrix_end;
	struct rusage myTime_mainAlg_end;
#endif
	node_type *rootO;
	node_type *rootP;

	pointList *O;
	pointList *P;

	int noOfDim;

	int i;

	int **preferenceO;
	int **preferenceP;

	assignmentSet *assignmentSetForO;
	assignmentSet *assignmentSetForP;

	int **rankP;

	float **matrix;

	int isConsistent;

	rtree_info *rtreeInfo;

	int noOfPointP, noOfPointO;

	assignment *aAssignment;
	twoData *aTwoData;

	debugFilePtr = (FILE *) fopen(debugFilename, "w");
	if (debugFilePtr == NULL)
	{
		printf("File %s cannot be written!\n", debugFilename);
		exit(0);
	}

	// statistics
	globalStat = statistics_new();

	// read configure
	printf("Reading configure file ...\n");
	aConfigure = readConfig(confFilename);

	//
	O = aConfigure->O;
	P = aConfigure->P;
	rtreeInfo = aConfigure->rtreeInfo;
	noOfDim = rtreeInfo->dim;

	noOfPointO = O->noOfPoint;
	noOfPointP = P->noOfPoint;

	aTwoData = twoData_new();
	aTwoData->O = pointList_clone(aConfigure->O, noOfDim);
	aTwoData->P = pointList_clone(aConfigure->P, noOfDim);

#ifndef WIN32
	getrusage(RUSAGE_SELF,&myTime_start);	
#endif	

	if (aConfigure->ALG == 1)	// our proposed alg
	{
		// build r-tree for O data
		printf("Building r-tree for O data ...\n");
		rootO = quota_buildRtree(O, rtreeInfo);

		//printf("Checking consistency of r-tree for O data ...\n");
		//isConsistent = rtree_checkConsistent(rootO, rtreeInfo);
		//printf("   consistent : %d\n", isConsistent);

#ifndef WIN32
	getrusage(RUSAGE_SELF,&myTime_buildRtreeO_end);	
#endif	

		// build r-tree for P data
		printf("Building r-tree for P data ...\n");
		rootP = quota_buildRtree(P, rtreeInfo);

		//printf("Checking consistency of r-tree for P data ...\n");
		//isConsistent = rtree_checkConsistent(rootP, rtreeInfo);
		//printf("   consistent : %d\n", isConsistent);

		// statistics
		statistics_rtreeNoOfNodeStorage(rootO, rootP, rtreeInfo);

#ifndef WIN32
	getrusage(RUSAGE_SELF,&myTime_buildRtreeP_end);	
#endif	

		// find stable assignment
		printf("Finding the assignment...\n");
		//aAssignment = findStableAssignment(aTwoData, rootO, rootP, aConfigure);
		aAssignment = spm_weightedChain(aTwoData, rootO, rootP, aConfigure);

#ifndef WIN32
	getrusage(RUSAGE_SELF,&myTime_mainAlg_end);	
#endif	

	}
	else if (aConfigure->ALG == 2)	// Gale-Shapley alg
	{
		// build the preference list
		printf("Building the preference list...\n");
		stable_constructTwoPreferenceList(aTwoData, &preferenceO, &preferenceP, noOfDim);

#ifndef WIN32
		getrusage(RUSAGE_SELF,&myTime_constructPrefList_end);	
#endif	
		// build the ranking list
		printf("Building the ranking list of P...\n");
		rankP = stable_constructOneRankList(preferenceP, noOfPointP, noOfPointO);

#ifndef WIN32
		getrusage(RUSAGE_SELF,&myTime_constructRankList_end);	
#endif	

		// free preferenceP
		printf("Freeing the preference list of P...\n");
		for (i = 0; i < noOfPointP; i++)
		{
			if (noOfPointO != 0)
			{
				free(preferenceP[i]);
			}
		}
		free(preferenceP);
//for (i = 0; i < 20; i++)		
{
//printf(" ~~~ %d ~~~~\n", i);
		// find stable assignment
		printf("Finding the assignment...\n");
		aAssignment = stable_GaleShapley(aTwoData, noOfDim, preferenceO, rankP);
}

#ifndef WIN32
		getrusage(RUSAGE_SELF,&myTime_mainAlg_end);	
#endif	

		// free rankList
		printf("Freeing...\n");
		for (i = 0; i < noOfPointP; i++)
		{
			if (noOfPointO != 0)
			{
				free(rankP[i]);
			}
		}
		free(rankP);

		// free preferenceO
		for (i = 0; i < noOfPointO; i++)
		{
			if (noOfPointP != 0)
			{
				free(preferenceO[i]);
			}
		}
		free(preferenceO);
	}
	else if (aConfigure->ALG == 3) // closest pair (fastest)
	{
		// build r-tree for O data
		printf("Building r-tree for O data ...\n");
		rootO = quota_buildRtree(O, rtreeInfo);

#ifndef WIN32
	getrusage(RUSAGE_SELF,&myTime_buildRtreeO_end);	
#endif	

		// build r-tree for P data
		printf("Building r-tree for P data ...\n");
		rootP = quota_buildRtree(P, rtreeInfo);

		// statistics
		statistics_rtreeNoOfNodeStorage(rootO, rootP, rtreeInfo);

#ifndef WIN32
	getrusage(RUSAGE_SELF,&myTime_buildRtreeP_end);	
#endif	

		// find stable assignment
		printf("Finding the assignment...\n");
		aAssignment = closestTree_findAssignment(aTwoData, rootO, rootP, aConfigure);

#ifndef WIN32
	getrusage(RUSAGE_SELF,&myTime_mainAlg_end);	
#endif			
	}
	else if (aConfigure->ALG == 4)	// closest pair (matrix)
	{
		// find a matrix
		printf("Creating the matrix...\n");
		matrix = closest_findDistanceMatrix(aTwoData, noOfDim);

#ifndef WIN32
		getrusage(RUSAGE_SELF,&myTime_constructMatrix_end);	
#endif	

		// main
		printf("Running the algorithm for cloest pair...\n");
		aAssignment = closest_main(aTwoData, noOfDim, matrix);

#ifndef WIN32
		getrusage(RUSAGE_SELF,&myTime_mainAlg_end);	
#endif		
		// free
		for (i = 0; i < noOfPointO; i++)
		{
			free(matrix[i]);
		}
		free(matrix);
	}

#ifndef WIN32

	timeFilePtr = (FILE *) fopen(timeFilename, "w");
	if (timeFilePtr == NULL)
	{
		printf("File %s cannot be written!\n", timeFilename);
		exit(0);
	}

	if (aConfigure->ALG == 1)
	{
		calculateExecutionTime(&(myTime_start), &(myTime_buildRtreeO_end), &userTime_buildRtreeO, &sysTime_buildRtreeO);
		calculateExecutionTime(&(myTime_buildRtreeO_end), &(myTime_buildRtreeP_end), &userTime_buildRtreeP, &sysTime_buildRtreeP);
		calculateExecutionTime(&(myTime_start), &(myTime_buildRtreeP_end), &userTime_totalBuildRTree, &sysTime_totalBuildRTree);

		calculateExecutionTime(&(myTime_buildRtreeP_end), &(myTime_mainAlg_end), &userTime_mainAlg, &sysTime_mainAlg);

		calculateExecutionTime(&(myTime_start), &(myTime_mainAlg_end), &userTime_allTime, &sysTime_allTime);

		fprintf(timeFilePtr, "%.10f\t%.10f\n", userTime_buildRtreeO, sysTime_buildRtreeO);
		fprintf(timeFilePtr, "%.10f\t%.10f\n", userTime_buildRtreeP, sysTime_buildRtreeP);
		fprintf(timeFilePtr, "%.10f\t%.10f\n", userTime_totalBuildRTree, sysTime_totalBuildRTree);

		fprintf(timeFilePtr, "\n");

		fprintf(timeFilePtr, "%.10f\t%.10f\n", userTime_mainAlg, sysTime_mainAlg);

		fprintf(timeFilePtr, "\n");

		fprintf(timeFilePtr, "%.10f\t%.10f\n", userTime_allTime, sysTime_allTime);
	}
	else if (aConfigure->ALG == 2)
	{
		calculateExecutionTime(&(myTime_start), &(myTime_constructPrefList_end), &userTime_constructPrefList, &sysTime_constructPrefList);
		calculateExecutionTime(&(myTime_constructPrefList_end), &(myTime_constructRankList_end), &userTime_constructRankList, &sysTime_constructRankList);
		calculateExecutionTime(&(myTime_start), &(myTime_constructRankList_end), &userTime_totalConstructPrefRankList, &sysTime_totalConstructPrefRankList);

		calculateExecutionTime(&(myTime_constructRankList_end), &(myTime_mainAlg_end), &userTime_mainAlg, &sysTime_mainAlg);

		calculateExecutionTime(&(myTime_start), &(myTime_mainAlg_end), &userTime_allTime, &sysTime_allTime);

		fprintf(timeFilePtr, "%.10f\t%.10f\n", userTime_constructPrefList, sysTime_constructPrefList);
		fprintf(timeFilePtr, "%.10f\t%.10f\n", userTime_constructRankList, sysTime_constructRankList);
		fprintf(timeFilePtr, "%.10f\t%.10f\n", userTime_totalConstructPrefRankList, sysTime_totalConstructPrefRankList);

		fprintf(timeFilePtr, "\n");

		fprintf(timeFilePtr, "%.10f\t%.10f\n", userTime_mainAlg, sysTime_mainAlg);

		fprintf(timeFilePtr, "\n");

		fprintf(timeFilePtr, "%.10f\t%.10f\n", userTime_allTime, sysTime_allTime);
	}
	else if (aConfigure->ALG == 3) 
	{
		calculateExecutionTime(&(myTime_start), &(myTime_buildRtreeO_end), &userTime_buildRtreeO, &sysTime_buildRtreeO);
		calculateExecutionTime(&(myTime_buildRtreeO_end), &(myTime_buildRtreeP_end), &userTime_buildRtreeP, &sysTime_buildRtreeP);
		calculateExecutionTime(&(myTime_start), &(myTime_buildRtreeP_end), &userTime_totalBuildRTree, &sysTime_totalBuildRTree);

		calculateExecutionTime(&(myTime_buildRtreeP_end), &(myTime_mainAlg_end), &userTime_mainAlg, &sysTime_mainAlg);

		calculateExecutionTime(&(myTime_start), &(myTime_mainAlg_end), &userTime_allTime, &sysTime_allTime);

		fprintf(timeFilePtr, "%.10f\t%.10f\n", userTime_buildRtreeO, sysTime_buildRtreeO);
		fprintf(timeFilePtr, "%.10f\t%.10f\n", userTime_buildRtreeP, sysTime_buildRtreeP);
		fprintf(timeFilePtr, "%.10f\t%.10f\n", userTime_totalBuildRTree, sysTime_totalBuildRTree);

		fprintf(timeFilePtr, "\n");

		fprintf(timeFilePtr, "%.10f\t%.10f\n", userTime_mainAlg, sysTime_mainAlg);

		fprintf(timeFilePtr, "\n");

		fprintf(timeFilePtr, "%.10f\t%.10f\n", userTime_allTime, sysTime_allTime);
	}
	else if (aConfigure->ALG == 4)
	{
		calculateExecutionTime(&(myTime_start), &(myTime_constructMatrix_end), &userTime_totalConstructMatrix, &sysTime_totalConstructMatrix);

		calculateExecutionTime(&(myTime_constructMatrix_end), &(myTime_mainAlg_end), &userTime_mainAlg, &sysTime_mainAlg);

		calculateExecutionTime(&(myTime_start), &(myTime_mainAlg_end), &userTime_allTime, &sysTime_allTime);

		fprintf(timeFilePtr, "%0 0\n");
		fprintf(timeFilePtr, "%0 0\n");
		fprintf(timeFilePtr, "%.10f\t%.10f\n", userTime_totalConstructMatrix, sysTime_totalConstructMatrix);

		fprintf(timeFilePtr, "\n");

		fprintf(timeFilePtr, "%.10f\t%.10f\n", userTime_mainAlg, sysTime_mainAlg);

		fprintf(timeFilePtr, "\n");

		fprintf(timeFilePtr, "%.10f\t%.10f\n", userTime_allTime, sysTime_allTime);
	}

	fclose(timeFilePtr);
#endif

	

	// print the assignment
	printf("Printing the assignment...\n");
	assignment_print(assignFilename, aAssignment);

	// transform to separate assignment
	printf("Transforming to separate assignments...\n");
	assignmentSet_transformFromFinalAssignment(aAssignment, noOfPointO, noOfPointP, &assignmentSetForO, &assignmentSetForP);

	// statistics
	printf("Summarizing statistics...\n");
	statistics_endMain(aConfigure, aTwoData, aAssignment, assignmentSetForO, assignmentSetForP);

	// print the debug information
	printf("Printing the debug information ....\n");
	statistics_print(debugFilePtr);

	// fclose debugFilePtr
	fclose(debugFilePtr);

	// statistics
	statistics_free(&globalStat);

	return 0;
}

*/