
/*
* synDataPreProcess contains some data-preprocessing utilities.
*/

#include "synDataPreProcess.h"


/*
* syn_data_process preprocesses the synthetic data generated by the generator--DaVisual code 1.0.
*/
void syn_data_process( )
{	
	int i, j, num, dim;
	float* obj1, *obj2;
	char filter[MAX_LINE_LENG];
	char file_name[MAX_FILE_NAME];
	char i_data_file[MAX_FILE_NAME];
	char o_data_file[MAX_FILE_NAME];
	FILE* c_fp, *i_fp, *o_fp;

	c_fp = fopen( SYN_CONFIG_FILE, "r");
	if( c_fp == NULL)
	{
		fprintf( stderr, "The syn_config.txt cannot be open.\n");
		exit( 0);
	}

	//Traversing all the input files;
	while( fscanf( c_fp, "%s%i%i", file_name, &num, &dim) != EOF)
	{
		obj1 = (float*)malloc( sizeof(float) * dim);
		obj2 = (float*)malloc( sizeof(float) * dim);

		strcpy( i_data_file, SYN_INPUT_FOLDER);
		strcat( i_data_file, file_name);			//input data file;

		strcpy( o_data_file, SYN_OUTPUT_FOLDER);
		strcat( o_data_file, file_name);			//output data file;

		if( (i_fp = fopen( i_data_file, "r")) == NULL ||
			(o_fp = fopen( o_data_file, "w")) == NULL)
		{
			fprintf( stderr, "The data file \"%s\" cannot be open.\n", file_name);
			exit(0);
		}
	
		//Filtering the description info in the input data file;
		for( i=0; i<FILTER_LINE_NUM; i++)
		{
			fgets( filter, MAX_LINE_LENG, i_fp);
		}

		//Writing the data info into the output data file;
		fprintf( o_fp, "%i\n", num);
		for( i=0; i<num; i++)
		{
			fscanf( i_fp, "%s", filter);
			//Reading the two bound points;
			for( j=0; j<dim; j++)
			{
				fscanf( i_fp, "%f", obj1+j);
			}
			for( j=0; j<dim; j++)
			{
				fscanf( i_fp, "%f", obj2+j);
			}
			
			//Writing the center point into the output file;
			for( j=0; j<dim; j++)
			{
				fprintf( o_fp, "%f  ", (obj1[j]+obj2[j])/2);
			}
			fprintf( o_fp, "\n");
		}

		//Releasing the two objects.
		free( obj1);
		free( obj2);
		fclose( i_fp);
		fclose( o_fp);
	}

	//Releasing the related resources.
	fclose( c_fp);
}

/*
* generate_weight generates the weight information for data objects.
*/
void generate_weight( )
{
	int i, num, lb, ub, ins_n;
	char file_name[MAX_FILE_NAME];
	char file_path[MAX_FILE_NAME];
	FILE* c_fp, *o_fp;

	if( (c_fp = fopen( SYN_WEI_CONFIG_FILE, "r")) == NULL)
	{
		fprintf( stderr, "The syn_wei_config file cannot be open.\n");
		exit(0);
	}

	srand( time(NULL));

	//Reading the configuration file;
	ins_n = 1;
	while( fscanf( c_fp, "%s%i%i%i", file_name, &num, &lb, &ub) != EOF)
	{
		if( ub < lb)
		{
			fprintf( stderr, "The inputed upper bound is not larger than the lower bound, \"%s\" is skipped.\n");
			continue;
		}
		else 
		{
			printf( "Generating weights: Case %i ...\n", ins_n++);

			sprintf( file_path, "%s%s", WEI_OUTPUT_FOLDER, file_name);
			if( (o_fp = fopen( file_path, "w")) == NULL)
			{
				fprintf( stderr, "The weight output file \"%s\" cannot be created.\n", file_name);
				exit(0);
			}

//			fprintf( o_fp, "%i\n", num);

			if( ub == lb)
			{
				for( i=0; i<num; i++)
					fprintf( o_fp, "%i\n", 1);
			}
			else	//ub > lb
			{
				for( i=0; i<num; i++)
				{
					fprintf( o_fp, "%i\n", rand()%(ub-lb) + lb);
				}
			}

			fclose( o_fp);
		}
	}

	//Releasing the related resources.
	fclose( c_fp);
}

/*
* combine_weight combines the data points info with the weight info.
*/
void combine_weight( )
{
	int i, j, num, dim, ins_n;
	obj_t obj_v;
	
	char i_points_file[MAX_FILE_NAME];
	char i_points_path[MAX_FILE_NAME];
	char i_weight_file[MAX_FILE_NAME];
	char i_weight_path[MAX_FILE_NAME];
	char combined_file[MAX_FILE_NAME];
	char des[MAX_LINE_LENG];
//	char combined_file[MAX_FILE_NAME];
	
	FILE* c_fp, *i_fp1, *i_fp2, *o_fp;
	
	if( (c_fp = fopen( COMBINE_CONFIG_FILE, "r")) == NULL)
	{
		fprintf( stderr, "The combine_config file cannot be open.\n");
		exit(0);
	}

	//Reading the points and weight info;
	ins_n = 1;
	while( fscanf( c_fp, "%s%s%s%s%s%i", des, i_points_file, des, i_weight_file, des, &dim) != EOF)
	{
		printf( "Combining weights: Case %i ...\n", ins_n++);

		sprintf( i_points_path, "%s%s", SYN_OUTPUT_FOLDER, i_points_file);
		sprintf( i_weight_path, "%s%s", SYN_OUTPUT_FOLDER, i_weight_file);
		sprintf( combined_file, "%s%s%s", SYN_OUTPUT_FOLDER, "Com-", i_points_file);

		if( ( i_fp1 = fopen( i_points_path, "r")) == NULL ||
			( i_fp2 = fopen( i_weight_path, "r")) == NULL ||
			( o_fp = fopen( combined_file, "w")) == NULL)
		{
			fprintf( stderr, "Cannot open the related file when dealing with the files \"%s\".\n", i_points_file);
			exit( 0);
		}

		fscanf( i_fp1, "%i", &num);
		obj_v.axis = (FLOAT*)malloc( sizeof(FLOAT) * dim);
		for( i=0; i<num; i++)
		{
			//Reading.
			for( j=0; j<dim; j++)
				fscanf( i_fp1, "%f", obj_v.axis+j);
			fscanf( i_fp2, "%f", &(obj_v.weight));

			//Writing.
			fprintf( o_fp, "%-8.0f  ", obj_v.weight);
			for( j=0; j<dim; j++)
				fprintf( o_fp, "%-8.3f  ", obj_v.axis[j]);
			fprintf( o_fp, "\n");
		}

		free( obj_v.axis);
		fclose( i_fp1);
		fclose( i_fp2);
		fclose( o_fp);
	}

	//Releasing the related resources.
	fclose( c_fp);
}


/*
* emp_combine_weight combines the data points info with the weight info.
*/
void emp_combine_weight( )
{
	int i, j, num, dim, ins_n;
	obj_t obj_v;
	
	char i_points_file[MAX_FILE_NAME];
	char i_points_path[MAX_FILE_NAME];
	char i_weight_file[MAX_FILE_NAME];
	char i_weight_path[MAX_FILE_NAME];
	char combined_file[MAX_FILE_NAME];
	char combined_path[MAX_FILE_NAME];
	char des[MAX_LINE_LENG];
//	char combined_file[MAX_FILE_NAME];
	
	FILE* c_fp, *i_fp1, *i_fp2, *o_fp;
	
	if( (c_fp = fopen( EMP_COM_CONFIG_FILE, "r")) == NULL)
	{
		fprintf( stderr, "The combine_config file cannot be open.\n");
		exit(0);
	}

	//Reading the points and weight info;
	ins_n = 1;
	while( fscanf( c_fp, "%s%i%s%i%s%s%s%s%s%s", des, &dim, des, &num, des, i_points_file, des, i_weight_file, des, combined_file) != EOF)
	{
		printf( "Combining weights: Case %i ...\n", ins_n++);

		sprintf( i_points_path, "%s%s", SYN_INPUT_FOLDER, i_points_file);
		sprintf( i_weight_path, "%s%s", WEI_OUTPUT_FOLDER, i_weight_file);
		sprintf( combined_path, "%s%s", SYN_OUTPUT_FOLDER, combined_file);

		if( ( i_fp1 = fopen( i_points_path, "r")) == NULL ||
			( i_fp2 = fopen( i_weight_path, "r")) == NULL ||
			( o_fp = fopen( combined_path, "w")) == NULL)
		{
			fprintf( stderr, "Cannot open the related file when dealing with the files \"%s\".\n", i_points_file);
			exit( 0);
		}

		//fscanf( i_fp1, "%i", &num);
		obj_v.axis = (FLOAT*)malloc( sizeof(FLOAT) * dim);
		for( i=0; i<num; i++)
		{
			//Reading.
			fscanf( i_fp1, "%s", des);
			for( j=0; j<dim; j++)
				fscanf( i_fp1, "%f", obj_v.axis+j);
			fscanf( i_fp2, "%f", &(obj_v.weight));

			//Writing.
			fprintf( o_fp, "%i: %-8.0f  ", i+1, obj_v.weight);
			for( j=0; j<dim; j++)
				fprintf( o_fp, "%-8.3f  ", obj_v.axis[j]);
			fprintf( o_fp, "\n");
		}

		free( obj_v.axis);
		fclose( i_fp1);
		fclose( i_fp2);
		fclose( o_fp);
	}

	//Releasing the related resources.
	fclose( c_fp);
}

/*
*	emp_real_data_norm normalizes the real data into [0,10000].
*/
void emp_real_data_norm( float min, float max)
{
	int i, j, n_t, dim, num;
	obj_t *obj_v, obj_min, obj_max;
	float offset, ratio;
	FILE *c_fp, *i_fp, *o_fp;

	char i_points_file[MAX_FILE_NAME];
	char i_points_path[MAX_FILE_NAME];
	char o_points_file[MAX_FILE_NAME];
	char o_points_path[MAX_FILE_NAME];
	char des[MAX_LINE_LENG];

	if( (c_fp = fopen( EMP_NORM_CONFIG_FILE, "r")) == NULL)
	{
		fprintf( stderr, "The combine_config file cannot be open.\n");
		exit(0);
	}
	
	//Reading the points and weight info;
	while( fscanf( c_fp, "%s%i%s%i%s%s%s%s", des, &dim, des, &num, des, i_points_file, des, o_points_file) != EOF)
	{
		sprintf( i_points_path, "%s%s", REAL_INPUT_FOLDER, i_points_file);
		sprintf( o_points_path, "%s%s", REAL_OUTPUT_FOLDER, o_points_file);

		if( ( i_fp = fopen( i_points_path, "r")) == NULL ||
			( o_fp = fopen( o_points_path, "w")) == NULL)
		{
			fprintf( stderr, "Cannot open the related file when dealing with the files \"%s\".\n", i_points_file);
			exit( 0);
		}

		obj_v = (obj_t*)malloc( sizeof( obj_t) * num);
		obj_min.axis = (FLOAT*)malloc( sizeof( FLOAT) * dim);
		obj_max.axis = (FLOAT*)malloc( sizeof( FLOAT) * dim);

		for( i=0; i<dim; i++)
		{
			obj_min.axis[i] = FLT_MAX;
			obj_max.axis[i] = -FLT_MAX;
		}

		//Reading.
		for( i=0; i<num; i++)
		{
			obj_v[i].axis = (FLOAT*)malloc( sizeof(FLOAT) * dim);
			for( j=0; j<dim; j++)
			{
				fscanf( i_fp, "%f%s", obj_v[i].axis+j, des);
				if( obj_v[i].axis[j] < obj_min.axis[j])
					obj_min.axis[j] = obj_v[i].axis[j];
				if( obj_v[i].axis[j] > obj_max.axis[j])
					obj_max.axis[j] = obj_v[i].axis[j];

				fscanf( i_fp, "%i", &n_t);
			}
			
		}
		
		//Normalizing.
		for( i=0; i<num; i++)
		{
			for( j=0; j<dim; j++)
			{
				offset = min - obj_min.axis[j];
				ratio = max/(obj_max.axis[j] + offset);

				obj_v[i].axis[j] += offset;
				obj_v[i].axis[j] *= ratio;
			}
		}

		//Outputing.
		for( i=0; i<num; i++)
		{
			//Writing.
			fprintf( o_fp, "%i	", i+1);
			for( j=0; j<dim; j++)
				fprintf( o_fp, "%f  ", obj_v[i].axis[j]);
			fprintf( o_fp, "\n");
		}

		for( i=0; i<num; i++)
			free( obj_v[i].axis);
		free( obj_v);
		free( obj_min.axis);
		free( obj_max.axis);
		fclose( i_fp);
		fclose( o_fp);
	}

	//Releasing the related the resource.
	fclose( c_fp);
}


/*
*	emp_real_data_gen generates the "specified" real datasets for the SPM-MM problem.
*/
void emp_real_data_gen( )
{
	int i, j, num, dim;
	obj_t obj_v;
	
	char i_points_file[MAX_FILE_NAME];
	char i_points_path[MAX_FILE_NAME];
	char i_weight_file[MAX_FILE_NAME];
	char i_weight_path[MAX_FILE_NAME];
	char combined_file[MAX_FILE_NAME];
	char combined_path[MAX_FILE_NAME];
	char des[MAX_LINE_LENG];
//	char combined_file[MAX_FILE_NAME];
	
	FILE* c_fp, *i_fp1, *i_fp2, *o_fp;
	
	if( (c_fp = fopen( EMP_COM_CONFIG_FILE, "r")) == NULL)
	{
		fprintf( stderr, "The combine_config file cannot be open.\n");
		exit(0);
	}

	//Reading the points and weight info;
	while( fscanf( c_fp, "%s%i%s%i%s%s%s%s%s%s", des, &dim, des, &num, des, i_points_file, des, i_weight_file, des, combined_file) != EOF)
	{
		sprintf( i_points_path, "%s%s", SYN_INPUT_FOLDER, i_points_file);
		sprintf( i_weight_path, "%s%s", SYN_INPUT_FOLDER, i_weight_file);
		sprintf( combined_path, "%s%s", SYN_OUTPUT_FOLDER, combined_file);

		if( ( i_fp1 = fopen( i_points_path, "r")) == NULL ||
			( i_fp2 = fopen( i_weight_path, "r")) == NULL ||
			( o_fp = fopen( combined_path, "w")) == NULL)
		{
			fprintf( stderr, "Cannot open the related file when dealing with the files \"%s\".\n", i_points_file);
			exit( 0);
		}

		//fscanf( i_fp1, "%i", &num);
		obj_v.axis = (FLOAT*)malloc( sizeof(FLOAT) * dim);
		for( i=0; i<num; i++)
		{
			//Reading.
			for( j=0; j<dim; j++)
				fscanf( i_fp1, "%f", obj_v.axis+j);
			fscanf( i_fp2, "%f", &(obj_v.weight));

			//Writing.
			fprintf( o_fp, "%i: %-8.0f  ", i+1, obj_v.weight);
			for( j=0; j<dim; j++)
				fprintf( o_fp, "%-8.3f  ", obj_v.axis[j]);
			fprintf( o_fp, "\n");
		}

		free( obj_v.axis);
		fclose( i_fp1);
		fclose( i_fp2);
		fclose( o_fp);
	}

	//Releasing the related resources.
	fclose( c_fp);
}